use std::sync::Arc;
use tokio::sync::{RwLock, mpsc};
use hyper::upgrade::Upgraded;
use hyper_util::rt::TokioIo;
use tokio_tungstenite::WebSocketStream;
use futures::StreamExt;

use crate::{
    protocol,
    util::pack::unpack_u16,
    HttpServiceArgs,
};

pub async fn handle_ws(ws:WebSocketStream<TokioIo<Upgraded>>, args:HttpServiceArgs) -> Result<(),()>
// Handle websocket connection.
//
{
    use tokio_tungstenite::tungstenite::protocol::Message;
    use protocol::{QRPacket, QRPacketData, code::*, packet::*};

    let conn_id :u32;

    let (sink, mut stream) = ws.split();
    let (tx, mut rx) = mpsc::channel::<QRPacket>(1);

    // Perform connection handshake with data system.
    //  - Provide system with connection/bus pairing.
    //  - Acquire connection id.
    //
    args.tx.send(QRPacket::new(0, QRPacketData::QConn {
        tx:tx.clone(),
        stream:Arc::new(RwLock::new(sink)),
    })).await.map_err(|_| ())?;

    if let Some(qr) = rx.recv().await {
        match &qr.data {
            QRPacketData::RConn => { conn_id = qr.id; }
            _ => { return Err(()); }
        }
    } else {
        return Err(());
    }

    drop(tx);
    drop(rx);

    // Decode client requests from websocket,
    // pass requests to data system,
    // and return responses to client.
    while match stream.next().await {
        Some(msg) => match msg {
            Ok(msg) => match msg {
                Message::Binary(data) => {
                    let mut index :usize = 0;
                    let code: u16 = unpack_u16(&data, &mut index);

                    match code {
                        CODE_HELLO => {
                            args.tx.send(
                                QRPacket::new(conn_id, QRPacketData::QHello)
                            ).await.ok();
                        }

                        CODE_REGISTER => match PacketRegister::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    QRPacket::new(conn_id, QRPacketData::QRegister(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_AUTH => match PacketAuth::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    QRPacket::new(conn_id, QRPacketData::QAuth(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_AUTH_RESUME => match PacketAuthResume::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    QRPacket::new(conn_id, QRPacketData::QAuthResume(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_AUTH_REVOKE => {
                            args.tx.send(
                                QRPacket::new(conn_id, QRPacketData::QAuthRevoke)
                            ).await.ok();
                        }

                        CODE_SESSION_LIST => match PacketSessionList::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    QRPacket::new(conn_id, QRPacketData::QSessionList(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        /*CODE_SESSION_JOIN => match PacketSessionCreate::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::QSessionCreate(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }*/

                        CODE_SESSION_VIEW => match PacketSessionView::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    QRPacket::new(conn_id, QRPacketData::QSessionView(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_SESSION_LEAVE => {
                            args.tx.send(
                                QRPacket::new(conn_id, QRPacketData::QSessionLeave)
                            ).await.ok();
                        }

                        CODE_SESSION_RETIRE => match PacketSessionRetire::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    QRPacket::new(conn_id, QRPacketData::QSessionResign(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_GAME_STATE => match PacketGameState::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    QRPacket::new(conn_id, QRPacketData::QGameState(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_GAME_MESSAGE => match PacketGameMessage::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    QRPacket::new(conn_id, QRPacketData::GameMessage(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_CHALLENGE => match PacketChallenge::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    QRPacket::new(conn_id, QRPacketData::QChallenge(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_CHALLENGE_ANSWER => match PacketChallengeAnswer::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    QRPacket::new(conn_id, QRPacketData::QChallengeAnswer(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_CHALLENGE_LIST => {
                            args.tx.send(
                                QRPacket::new(conn_id, QRPacketData::QChallengeList)
                            ).await.ok();
                        }

                        CODE_USER_LIST => {
                            args.tx.send(
                                QRPacket::new(conn_id, QRPacketData::QUserList)
                            ).await.ok();
                        }

                        CODE_USER_PROFILE => match PacketUserProfile::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.tx.send(
                                    QRPacket::new(conn_id, QRPacketData::QUserProfile(packet))
                                ).await.ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_INVITE_LIST => {
                            args.tx.send(
                                QRPacket::new(conn_id, QRPacketData::QInviteList)
                            ).await.ok();
                        }

                        CODE_INVITE_ACQUIRE => {
                            args.tx.send(
                                QRPacket::new(conn_id, QRPacketData::QInviteAcquire)
                            ).await.ok();
                        }

                        _ => { }
                    }
                    true
                }
                
                Message::Close(_) => { true }

                _ => {
                    println!("notice: received unexpected websocket data.");
                    true
                }
            }
            Err(e) => {
                println!("WSS error: {}", e.to_string());
                false
            },
        }
        None => false,
    } { }

    args.tx.send(QRPacket::new(conn_id, QRPacketData::QDisconn)).await.ok();
    Ok(())
}
